;;;; csi.scm - Simple interpreter stub for CHICKEN
;
; Copyright (c) 2000-2003, Felix L. Winkelmann
; All rights reserved.
;
; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
; conditions are met:
;
;   Redistributions of source code must retain the above copyright notice, this list of conditions and the following
;     disclaimer. 
;   Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
;     disclaimer in the documentation and/or other materials provided with the distribution. 
;   Neither the name of the author nor the names of its contributors may be used to endorse or promote
;     products derived from this software without specific prior written permission. 
;
; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
; OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
; AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR
; CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
; CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
; SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
; THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
; OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
; POSSIBILITY OF SUCH DAMAGE.
;
; Send bugs, suggestions and ideas to: 
;
; felix@call-with-current-continuation.org
;
; Felix L. Winkelmann
; Steinweg 1A
; 37130 Gleichen, OT Weissenborn
; Germany


(declare
  (uses extras srfi-1 syntax-case srfi-4 srfi-18 srfi-37 format match-support tinyclos regex script-utils)
  (usual-integrations)
  (disable-interrupts)
  (foreign-declare #<<EOF
#ifdef C_USE_C_DEFAULTS
# include "c_defaults.h"
#elif !defined(C_INSTALL_HOME)
# define C_INSTALL_HOME NULL
#endif

#if defined(_MSC_VER) && defined(_WIN32)
# include <direct.h>
#else
# define _getcwd(buf, len)       NULL
#endif

#ifdef __DJGPP__
# include <unistd.h>
#endif
EOF
) )

(include "moremacros")
(include "match")

#{csi 
  max-history-count print-usage print-banner write-to-echo-port
  run hexdump delq transcript-port
  trace-hl-macros parse-option-string chop-separator lookup-script-file
  report describe dump hexdump bytevector-data get-config
  advised-procedures do-advise do-unadvise
  trace-indent trace-indent-level traced-procedures traced-procedure-entry traced-procedure-exit}

(declare
  (hide parse-option-string trace-hl-macros bytevector-data transcript-port slibpath
	write-to-echo-port trace-indent setup-usage) )


;;; Parameters:

(define-constant init-file ".csirc")
(define-constant max-history-count 16)

(set! ##sys#repl-print-length-limit 512)

(define-foreign-variable installation-home c-string "C_INSTALL_HOME")

(define slibpath (getenv "SCHEME_LIBRARY_PATH"))


;;; Declare used units as provided:

(provide
 'extras 'srfi-1 'syntax-case 'srfi-4
 'srfi-18 'format 'tinyclos 'regex
 'script-utils)

(when (memq #:posix (features)) (provide 'posix))
(when (memq #:tcp (features)) (provide 'tcp))


;;; Print all sorts of information:

(define (print-usage)
  (display #<#EOF
Usage: csi {FILENAME | OPTION}

  where OPTION may be one of the following:

    -help                       display this text and exit
    -version                    display version and exit
    -case-insensitive           enable case-insensitive reading
    -eval EXPRESSIONS           evaluate given expressions
    -feature SYMBOL             register feature identifier
    -quiet                      do not print banner
    -no-init                    do not load initialization file `~/#{init-file}'
    -no-feature SYMBOL          unregister feature identifier
    -batch                      terminate after command-line processing
    -hygienic                   install SYNTAX-CASE macro package
    -no-warnings                disable all warnings
    -keyword-style STYLE        enable alternative keyword-syntax (prefix or suffix)
    -r5rs                       equivalent to `-hygienic -strict'
    -script PATHNAME            use interpreter for shell scripts
    -script-meta PATHNAME       `-script' with extra options
    -srfi-7                     process source files as SRFI-7 config. language
    -strict                     disable non-standard macros and symbol syntax
    -strict-srfi-0              disable non-standard macros except `cond-expand'
    -strict-reader              disable non-standard read syntax
    -strict-letrec              enable fully R5RS conformant `letrec'
    -setup [OPTION | EXTENSION] unpack extension ("egg") and run setup script
    -setup-help                 show list of available setup options
    -visit FILENAME             visit (macroexpand) source file
    -slib                       enable SLIB support (needs additional files)
    --                          ignore all following options

EOF
) )

(define (setup-usage)
  (print #<<EOF
usage: csi -setup [EXTENSION] [MODE]

  -i  --install        install extension, build if needed
  -b  --build          build extension, extract if needed
  -w  --wrap           create wrapped package
  -u  --uninstall      uninstall extension
EOF
  )
  (exit 0) )

(define print-banner
  (let ([config 
	 (let ([bp (build-platform)]
	       [st (software-type)]
	       [sv (software-version)]
	       [mt (machine-type)] )
	   (define (str x)
	     (if (eq? x 'unknown)
		 ""
		 (string-append (##sys#slot x 1) "-") ) )
	   (string-append (str sv) (str st) (str bp) (##sys#slot mt 1)) ) ] )
    (lambda ()
      (printf "; This is the CHICKEN interpreter.~%~
               ; Version ~S, Build ~S - ~A~%~
               ; (c)2000-2003 Felix L. Winkelmann~%"
	      build-version build-number config) ) ) )


;;; Reader for she-bang comment:

(set! ##sys#user-read-hook
  (let ([read-char read-char]
	[read read]
	[old-hook ##sys#user-read-hook] )
    (lambda (char port)
      (if (char=? char #\!)
	  (do ((c (read-char port) (read-char port)))
	      ((char=? c #\newline) (read port))
	    (when (eof-object? c) (##sys#error "unexpected end of file")) )
	  (old-hook char port) ) ) ) )


;;; Chop terminating separator from pathname:

(define chop-separator 
  (let ([pds pathname-directory-separator]
	[substring substring] )
    (lambda (str)
      (let ([len (sub1 (##sys#size str))])
	(if (and (fx> len 0) (char=? (string-ref str len) pds))
	    (substring str 0 len)
	    str) ) ) ) )


;;; Find script in PATH (only used for Windows/DOS):

(define @ #f)

(define lookup-script-file 
  (let* ([buf (make-string 256)]
	 [_getcwd (foreign-lambda nonnull-c-string "_getcwd" pointer int)] )
    (define (addext name)
      (if (file-exists? name)
	  name
	  (let ([n2 (string-append name ".bat")])
	    (and (file-exists? n2) n2) ) ) )
    (define (string-index proc str1)
      (let ((len (##sys#size str1)))
	(let loop ((i 0))
	  (cond ((fx>= i len) #f)
		((proc (##core#inline "C_subchar" str1 i)) i)
		(else (loop (fx+ i 1))) ) ) ) )
    (lambda (name)
      (let ([path (getenv "PATH")])
	(and (> (##sys#size name) 0)
	     (cond [(char=? pathname-directory-separator (string-ref name 0)) (addext name)]
		   [(string-index (lambda (c) (char=? c pathname-directory-separator)) name)
		    (and-let* ([p (_getcwd buf 256)])
		      (addext (string-append (chop-separator p) (string pathname-directory-separator) name)) ) ]
		   [(addext name)]
		   [else
		    (let ([name2 (string-append (string pathname-directory-separator) name)])
		      (let loop ([ps (string-split path ";")])
			(and (pair? ps)
			     (let ([name2 (string-append (chop-separator (##sys#slot ps 0)) name2)])
			       (or (addext name2)
				   (loop (##sys#slot ps 1)) ) ) ) ) ) ] ) ) ) ) ) )
				   

;;; Add some 'comma'-commands:

(set! ##sys#repl-eval-hook
  (let ((eval eval)
	(load-noisily load-noisily)
	(read read)
	(length length)
	(display display)
	(write write)
	(printf printf)
	(integer? integer?)
	(second second)
	(butlast butlast)
	(take take)
	(list-ref list-ref)
	(values values)
	(history-list '())
	(history-count 0) 
	(void void) )
    (lambda (form)

      (define (filename x)
	(cond ((string? x) x)
	      ((symbol? x) (##sys#symbol->string x)) 
	      (else (##sys#error "argument is not a valid filename" x)) ) )

      (define (add-h form results)
	(cond ((fx>= history-count max-history-count)
	       (set! history-list 
		 (cons (cons form results) 
		       (take history-list (fx- history-count 1)) ) ) )
	      (else
	       (set! history-list (cons (cons form results) history-list))
	       (set! history-count (fx+ history-count 1)) ) ) )

      (define (ref-h index)
	(let ((i (fx- (inexact->exact index) 1)))
	  (if (and (fx>= i 0) (fx< i (length history-list)))
	      (list-ref history-list i) 
	      (##sys#error "history entry index out of range" index) ) ) )
	  
      (set! trace-indent-level 0)
      (set! $ (lambda i
		(let ((index (if (pair? i) (inexact->exact (##sys#slot i 0)) 1)))
		  (if (integer? index)
		      (car (ref-h index))
		      (##sys#error "invalid history entry index" index) ) ) ) )
      (set! & (lambda i
		(let ((index (if (pair? i) (inexact->exact (##sys#slot i 0)) 1)))
		  (if (integer? index)
		      (apply values (cdr (ref-h index)))
		      (##sys#error "invalid history entry index" index) ) ) ) )
      (write-to-echo-port form)
      (cond ((eof-object? form) (exit))
	    ((and (pair? form)
		  (eq? 'unquote (##sys#slot form 0)) )
	     (let ((cmd (second form)))
	       (case cmd
		 ((x)
		  (let ([x (read)])
		    (add-h `(pretty-print (##sys#secondary-macroexpand (macroexpand ',x))) (list (void)))
		    (pretty-print (macroexpand x))
		    (void) ) )
		 ((p)
		  (let* ([x (read)]
			 [xe (eval x)] )
		    (add-h `(pretty-print (eval ',x)) (list (void)))
		    (pretty-print xe)
		    (void) ) )
		 ((d)
		  (let* ([x (read)]
			 [xe (eval x)] )
		    (describe xe) ) )
		 ((du)
		  (let* ([x (read)]
			 [xe (eval x)] )
		    (dump xe) ) )
		 ((dur)
		  (let* ([x (read)]
			 [n (read)]
			 [xe (eval x)] 
			 [xn (eval n)] )
		    (dump xe xn) ) )
		 ((r) (report))
		 ((q) (exit))
		 ((l)
		  (let ((fn (filename (read))))
		    (add-h `(load ,fn) (list (void)))
		    (load fn)
		    (void) ) )
		 ((l7)
		  (let ((fn (filename (read))))
		    (add-h `(load-srfi-7-program ,fn) (list (void)))
		    (load-srfi-7-program fn)
		    (void) ) )
		 ((ln) 
		  (let ((fn (filename (read))))
		    (add-h `(load-noisily ,fn) (list (void)))
		    (load-noisily fn)
		    (void) ) )
		 ((t)
		  (let ((x (read)))
		    (receive rs (time (eval x))
		      (add-h `(time (eval ',x)) rs)
		      (apply values rs) ) ) )
		 ((s)
		  (let* ((x (read))
			 (str (if (symbol? x) (##sys#symbol->string x) x))
			 (r (system str)) )
		    (add-h `(system ,str) (list r))
		    r) )
		 ((h)
		  (do ((i 1 (fx+ i 1))
		       (hl history-list (##sys#slot hl 1)) )
		      ((eq? hl '()))
		    (let ((entry (##sys#slot hl 0)))
		      (printf "~s~a: ~s" i (if (fx< i 10) #\space "") (car entry))
		      (if (not (equal? (list (void)) (cdr entry)))
			  (begin
			    (display " ->")
			    (for-each 
			     (lambda (x)
			       (##sys#with-print-length-limit
				60
				(lambda ()
				  (display #\space)
				  (write x) ) ) )
			     (cdr entry) ) ) )
		      (newline) ) )
		  (void) )
		 ((?)
		  (display 
		   "Toplevel commands:

 ,?            Show this text
 ,h            List history of last commands
 ,NUM          Evaluate NUMth command in history-list
 ,p EXP        Pretty print evaluated expression EXP
 ,d EXP        Describe result of evaluated expression EXP
 ,du EXP       Dump data of expression EXP
 ,dur EXP N    Dump range
 ,q            Quit interpreter
 ,l FILENAME   Load file with given filename (may be symbol or string)
 ,ln FILENAME  Load file and print result of each top-level expression
 ,l7 FILENAME  Process file as SRFI-7 configuration language
 ,r            Show system information
 ,s STRING     Execute shell-command
 ,t EXP        Evaluate form and print elapsed time
 ,x EXP        Pretty print macroexpanded expression EXP\n")
		  (void) )
		 (else
		  (cond ((fixnum? cmd)
			 (let* ((item (ref-h cmd))
				(exp (car item)) )
			   (receive rs (eval exp)
			     (add-h exp rs)
			     (apply values rs) ) ) )
			(else
			 (printf "Undefined toplevel command ~s - enter `,?' for help~%" form) 
			 (void) ) ) ) ) ) )
	    (else
	     (receive rs (eval form)
	       (add-h form rs) 
	       (apply values rs) ) ) ) ) ) )

(define transcript-port #f)

(define write-to-echo-port
  (let ([write write]
	[newline newline] )
    (lambda (x)
      (when transcript-port 
	(write x transcript-port)
	(newline transcript-port) ) ) ) )

(define transcript-on 
  (let ([close-output-port close-output-port]
	[open-output-file open-output-file] )
    (lambda (fname)
      (##sys#check-string fname 'transcript-on)
      (when transcript-port (close-output-port transcript-port))
      (set! transcript-port (open-output-file fname)) ) ) )

(define transcript-off
  (let ([close-output-port close-output-port])
    (lambda ()
      (when transcript-port (close-output-port transcript-port))
      (set! transcript-port #f) ) ) )


;;; Display prompt if stdin is a terminal:

(set! ##sys#read-prompt-hook
  (let ([display display])
    (lambda ()
      (when (or (##sys#fudge 12) (##sys#tty-port? (current-input-port)))
	(display ">>> ") 
	(flush-output) ) ) ) )


;;; Tracing/advise:

(define (delq x lst)
  (let loop ([lst lst])
    (if (null? lst)
	'()
	(let ([y (car lst)])
	  (if (eq? x y)
	      (cdr lst)
	      (cons y (loop (cdr lst))) ) ) ) ) )

(define trace-indent-level 0)
(define traced-procedures '())
(define advised-procedures '())

(define trace-indent
  (lambda ()
    (write-char #\|)
    (do ((i trace-indent-level (sub1 i)))
	((<= i 0))
      (write-char #\space) ) ) )

(define traced-procedure-entry
  (lambda (name args)
    (trace-indent)
    (set! trace-indent-level (add1 trace-indent-level))
    (write (cons name args))
    (newline) 
    (flush-output) ) )

(define traced-procedure-exit
  (lambda (name results)
    (set! trace-indent-level (sub1 trace-indent-level))
    (trace-indent)
    (write name)
    (display " -> ")
    (for-each
     (lambda (x)
       (write x)
       (write-char #\space) )
     results)
    (newline) 
    (flush-output) ) )

(define (do-advise name old)
  (unless (assq name advised-procedures)
    (when (not (procedure? old))
      (##sys#signal-hook #:type-error 'advise "can not advise non-procedure" old) )
    (set! advised-procedures (cons (cons name old) advised-procedures)) ) )

(define (do-unadvise name prev)
  (let ([a (assq name advised-procedures)])
    (if a
	(begin
	  (set! advised-procedures (delq a advised-procedures))
	  (##sys#slot a 1) )
	(begin
	  (printf "Procedure `~S' was never advised.~%" name)
	  prev) ) ) )

(define-macro (trace . names)
  `(begin
     ,@(map (lambda (s)
	      (let ((name (gensym))
		    (old (gensym)) )
		`(let* ((,name ',s)
			(,old ,s) )
		   (if (not (procedure? ,old)) (##sys#error 'trace "can not trace non-procedure" ,old))
		   (set! traced-procedures (cons (cons ,name ,old) traced-procedures))
		   (set! ,s (lambda args
			      (traced-procedure-entry ,name args)
			      (call-with-values (lambda () (apply ,old args))
				(lambda results
				  (traced-procedure-exit ,name results) 
				  (apply values results) ) ) ) ) ) ) )
	    names) ) )

(define-macro (untrace . names)
  `(begin
     ,@(map (lambda (s)
	      (let ((var (gensym))
		    (old (gensym)) )
		`(let* ((,var ',s)
			(,old (assq ,var traced-procedures)) )
		   (if ,old
		       (set! ,s (cdr ,old))
		       (printf "Procedure `~s' was not traced.~%" ,var) )
		   (set! traced-procedures (delq ,old traced-procedures) ) ) ) )
	    names) ) )

(define-macro (advise name mode proc)
  (let ([tmp (gensym)]
	[old (gensym)] )
    `(let* ([,tmp ,proc]
	    [,old ,name] )
       (do-advise ',name ,old)
       (set! ,name
	 ,(case mode
	    [(before)
	     `(lambda args
		(apply ,tmp args)
		(apply ,old args) ) ]
	    [(after)
	     `(lambda args
		(call-with-values
		    (lambda () (apply ,old args))
		  (lambda results
		    (apply ,tmp results)
		    (apply values results) ) ) ) ]
	    [(around)
	     `(lambda args
		(apply ,tmp ,old args) ) ]
	    [else (##sys#error 'advise "invalid mode" ',mode)] ) ) ) ) )

(define-macro (unadvise . names)
  `(begin
     ,@(map (lambda (n) 
	      `(set! ,n (do-unadvise ',n ,n) ) )
	    names) ) )

(define trace-hl-macros #<<EOF
(begin

(define-syntax trace
  (lambda (x)
    (syntax-case x ()
      [(_) (syntax (##core#undefined))]
      [(_ name ...)
       (with-syntax ([(tmp ...) (generate-temporaries (syntax (name ...)))])
	 (syntax
	  (let ([tmp name] ...)
	    (if (not (procedure? tmp)) (##sys#error 'trace "can not trace non-procedure" tmp)) ...
	    (set! ##csi#traced-procedures (cons (cons 'name tmp) ##csi#traced-procedures)) ...
	    (set! name (lambda args
			 (##csi#traced-procedure-entry 'name args)
			 (##sys#call-with-values (lambda () (##sys#apply tmp args))
			   (lambda results
			     (##csi#traced-procedure-exit 'name results)
			     (##sys#apply ##sys#values results) ) ) ) )
	    ...) ) ) ] ) ) )

(define-syntax untrace
  (lambda (x)
    (syntax-case x ()
      [(_) (syntax (##core#undefined))]
      [(_ name ...)
       (with-syntax ([(tmp ...) (generate-temporaries (syntax (name ...)))])
	 (syntax
	  (let ([tmp (assq 'name ##csi#traced-procedures)] ...)
	    (if tmp (set! name (cdr tmp))) ...
	    (set! ##csi#traced-procedures (##csi#delq tmp ##csi#traced-procedures)) ...) ) ) ] ) ) )

(define-syntax advise
  (syntax-rules (before after around)
    [(_ "1" name before tmp old)
     (set! name (lambda args (apply tmp args) (apply old args))) ]
    [(_ "1" name after tmp old)
     (set! name
       (lambda args
	 (call-with-values (lambda () (apply old args))
	   (lambda results 
	     (apply tmp results)
	     (apply values results) ) ) ) ) ]
    [(_ "1" name around tmp old)
     (set! name (lambda args (apply tmp old args))) ]
    [(_ name mode proc)
     (let ([tmp proc]
	   [old name] )
       (##csi#do-advise 'name old)
       (advise "1" name mode tmp old) ) ] ) )

(define-syntax unadvise
  (syntax-rules ()
    [(_ name ...)
     (begin
       (set! name (##csi#do-unadvise 'name name)) ...) ] ) )

)
EOF
)


;;; Parse options from string:

(define (parse-option-string str)
  (with-input-from-string str
    (lambda ()
      (map (lambda (o)
	     (if (string? o)
		 o
		 (sprintf "~S" o) ) )
	   (handle-exceptions ex (##sys#error "invalid option syntax" str)
	     (read-file) ) ) ) ) )


;;; Print status information:

(define report
  (let ([format format]
	[with-output-to-port with-output-to-port] 
	[current-output-port current-output-port] )
    (lambda port
      (with-output-to-port (if (pair? port) (car port) (current-output-port))
	(lambda ()
	  (let ([sinfo (##sys#symbol-table-info)]
		[minfo (##sys#memory-info)] )
	    (define (shorten n) (/ (truncate (* n 100)) 100))
	    (let-values ([(sysname nodename release version machine) 
			  (cond-expand [posix (system-information)]
				       [else (values "<unknown>" "" "" "" "")] ) ] )
	      (format #t "Features:~%~{  ~A~%~}~%~
                   Host system information:~%  ~A ~A ~A ~A ~A~%~%~
                   Machine type:    \t~A ~A~%~
                   Software type:   \t~A~%~
                   Software version:\t~A~%~
                   Build platform:  \t~A~%~%~
                   Current working dir:\t~A~%~
                   Include path:~%~{  ~A~%~}~%~
                   Symbol-table load:\t~S~%  ~
                     Avg bucket length:\t~S~%  ~
                     Total symbols:\t~S~%~%~
                   Memory:\t~S heap~A, ~S stack, stack grows ~A~%~
                   Bytes free in heap:\t~S~%"
		      (sort (features) (lambda (x y) (string<? (symbol->string x) (symbol->string y))))
		      sysname nodename release version machine
		      (machine-type)
		      (if (##sys#fudge 3) "(64-bit)" "")
		      (software-type)
		      (software-version)
		      (build-platform)
		      (cond-expand [posix (current-directory)]
				   [else "<unknown>"] )
		      ##sys#include-pathnames
		      (shorten (vector-ref sinfo 0))
		      (shorten (vector-ref sinfo 1))
		      (vector-ref sinfo 2)
		      (vector-ref minfo 0)
		      (if (##sys#fudge 17) " (fixed)" "")
		      (vector-ref minfo 1)
		      (if (= 1 (##sys#fudge 18)) "downward" "upward")
		      (gc) )
	      (newline)
	      (when (##sys#fudge 14) (display "interrupts are enabled\n"))
	      (when (##sys#fudge 15) (display "symbol gc is enabled\n")) 
	      (##core#undefined) ) ) ) ) ) ) )


;;; Describe & dump:

(define bytevector-data
  `((u8vector "vector of unsigned bytes" ,u8vector-length ,u8vector-ref)
    (s8vector "vector of signed bytes" ,s8vector-length ,s8vector-ref)
    (u16vector "vector of unsigned 16-bit words" ,u16vector-length ,u16vector-ref)
    (s16vector "vector of signed 16-bit words" ,s16vector-length ,s16vector-ref)
    (u32vector "vector of unsigned 32-bit words" ,u32vector-length ,u32vector-ref)
    (s32vector "vector of signed 32-bit words" ,s32vector-length ,s32vector-ref)
    (f32vector "vector of 32-bit floats" ,f32vector-length ,f32vector-ref)
    (f64vector "vector of 64-bit floats" ,f64vector-length ,f64vector-ref) ) )

(define-constant max-describe-lines 40)

(define describe
  (let ([sprintf sprintf]
	[printf printf] 
	[fprintf fprintf]
	[length length]
	[list-ref list-ref]
	[string-ref string-ref]
	[describe-object describe-object]
	[instance? instance?]
	[newline newline]
	[hash-table-for-each hash-table-for-each] 
	[void void] )
    (lambda (x . port)
      (let ([out (:optional port ##sys#standard-output)])

	(define (descseq name plen pref start)
	  (##sys#call-with-current-continuation
	   (lambda (return)
	     (let ([len (fx- (plen x) start)])
	       (when name (fprintf out "~A of length ~S~%" name len))
	       (do ([i 0 (fx+ i 1)])
		   ((fx>= i len))
		 (when (fx>= i max-describe-lines)
		   (fprintf out "~% (~A elements not displayed)~%" (fx- len i))
		   (return #f) )
		 (fprintf out " ~S: ~S~%" i (pref x (fx+ start i))) ) ) ) ) )

	(when (##sys#permanent? x)
	  (fprintf out "statically allocated (~X) " (##sys#block-address x)) )
	(cond [(char? x)
	       (let ([code (char->integer x)])
		 (fprintf out "character ~S, code: ~S, #x~X, #o~O~%" x code code code) ) ]
	      [(eq? x #t) (fprintf out "boolean true~%")]
	      [(eq? x #f) (fprintf out "boolean false~%")]
	      [(null? x) (fprintf out "empty list~%")]
	      [(eof-object? x) (fprintf out "end-of-file~%")]
	      [(eq? (void) x) (fprintf out "unspecified object~%")]
	      [(fixnum? x)
	       (fprintf out "exact integer ~S, #x~X, #o~O, #b~B" x x x x)
	       (when (fx< x 256) (fprintf out ", character ~S" (integer->char x)))
	       (newline) ]
	      [(eq? x (##sys#slot '##sys#arbitrary-unbound-symbol 0))
	       (fprintf out "unbound value~%") ]
	      [(number? x) (fprintf out "inexact number ~S~%" x)]
	      [(string? x) (descseq "string" ##sys#size string-ref 0)]
	      [(vector? x) (descseq "vector" ##sys#size ##sys#slot 0)]
	      [(symbol? x)
	       (unless (##sys#symbol-has-toplevel-binding? x) (display "unbound " out))
	       (when (and (symbol? x) (fx= 0 (##sys#byte (##sys#slot x 1) 0)))
		 (display "keyword " out) )
	       (fprintf out "symbol with name ~S~%" (##sys#symbol->string x)) ]
	      [(list? x) (descseq "list" length list-ref 0)]
	      [(pair? x) (fprintf out "pair with car ~S and cdr ~S~%" (car x) (cdr x))]
	      [(procedure? x)
	       (let ([len (##sys#size x)])
		 (if (and (> len 3)
			  (eq? ##tinyclos#entity-tag (##sys#slot x (fx- len 1))) )
		     (describe-object x out)
		     (descseq 
		      (sprintf "procedure with code pointer ~X" (##sys#peek-unsigned-integer x 0))
		      ##sys#size ##sys#slot 1) ) ) ]
	      [(port? x)
	       (fprintf out
			"~A port of type ~A with name ~S and file pointer ~X~%"
			(if (##sys#slot x 1) "input" "output")
			(##sys#slot x 7)
			(##sys#slot x 3)
			(##sys#peek-unsigned-integer x 0) ) ]
	      [(instance? x)
	       (describe-object x out) ]
	      [(##sys#locative? x)
	       (fprintf out "locative~%  pointer ~X~%  index ~A~%  type ~A~%"
			(##sys#peek-unsigned-integer x 0)
			(##sys#slot x 1)
			(case (##sys#slot x 2) 
			  [(0) "slot"]
			  [(1) "char"]
			  [(2) "u8vector"]
			  [(3) "s8vector"]
			  [(4) "u16vector"]
			  [(5) "s16vector"]
			  [(6) "u32vector"]
			  [(7) "s32vector"]
			  [(8) "f32vector"]
			  [(9) "f64vector"] ) ) ]
	      [(##sys#pointer? x) (fprintf out "machine pointer ~X~%" (##sys#peek-unsigned-integer x 0))]
	      [(##sys#bytevector? x)
	       (let ([len (##sys#size x)])
		 (fprintf out "byte vector of size ~S:~%" len)
		 (hexdump x len ##sys#byte out) ) ]
	      [(##sys#structure? x 'hash-table)
	       (let ([test
		      (switch (##sys#slot x 3)
			[eq? 'eq?]
			[eqv? eqv?]
			[equal? 'equal?] 
			[else "<unknown>"] ) ] )
		 (fprintf out "hash-table with ~S elements and comparison procedure ~A:~%" (##sys#slot x 2) test)
		 (hash-table-for-each
		  (lambda (k v) (fprintf out " ~S\t-> ~S~%" k v))
		  x) ) ]
	      [(##sys#structure? x 'environment)
	       (if (##sys#slot x 1)
		   (fprintf out "environment~%") 
		   (fprintf out "the interaction environment~%") ) ]
	      [(##sys#structure? x 'array)
	       (let ([shp (##sys#slot x 3)]
		     [vec (##sys#slot x 1)] )
		 (fprintf out "array of rank ~S with ~A elements~% shape: " (fx/ (##sys#size shp) 2) (##sys#size vec))
		 (let loop ([s (vector->list shp)])
		   (unless (null? s)
		     (let ([next (##sys#slot s 1)])
		       (fprintf out " ~A..~A " (##sys#slot s 0) (##sys#slot next 0))
		       (loop (##sys#slot next 1)) ) ) )
		 (newline out) ) ]
	      [(##sys#generic-structure? x)
	       (let ([st (##sys#slot x 0)])
		 (let ([data (assq st bytevector-data)] )
		   (if data
		       (apply descseq (append (cdr data) (list 0)))
		       (begin
			 (fprintf out "structure of type ~S:~%" (##sys#slot x 0))
			 (descseq #f ##sys#size ##sys#slot 1) ) ) ) ) ]
	      [else (fprintf out "unknown object~%")] )
	(void) ) ) ) )


;;; Display hexdump:

(define dump
  (lambda (x . len-out)
    (let-optionals len-out
	([len #f]
	 [out ##sys#standard-output] )
      (define (bestlen n) (if len (min len n) n))
      (cond [(##sys#immediate? x) (##sys#error 'dump "can not dump immediate object" x)]
	    [(##sys#bytevector? x) (hexdump x (bestlen (##sys#size x)) ##sys#byte out)]
	    [(string? x) (hexdump x (bestlen (##sys#size x)) ##sys#byte out)]
	    [(and (not (##sys#immediate? x)) (##sys#pointer? x))
	     (hexdump x 32 ##sys#peek-byte out) ]
	    [(and (##sys#generic-structure? x) (assq (##sys#slot x 0) bytevector-data))
	     (let ([bv (##sys#slot x 1)])
	       (hexdump bv (bestlen (##sys#size bv)) ##sys#byte out) ) ]
	    [else (##sys#error 'dump "can not dump object" x)] ) ) ) )

(define hexdump
  (let ([display display]
	[string-append string-append]
	[make-string make-string]
	[write-char write-char] )
    (lambda (bv len ref out)

      (define (justify n m base lead)
	(let* ([s (number->string n base)]
	       [len (##sys#size s)] )
	  (if (fx< len m)
	      (string-append (make-string (fx- m len) lead) s)
	      s) ) )

      (do ([a 0 (fx+ a 16)])
	  ((fx>= a len))
	(display (justify a 4 10 #\space) out)
	(write-char #\: out)
	(do ([j 0 (fx+ j 1)]
	     [a a (fx+ a 1)] )
	    ((or (fx>= j 16) (fx>= a len))
	     (and-let* ([(fx>= a len)]
			[o (fxmod len 16)]
			[(not (fx= o 0))] )
	       (do ([k (fx- 16 o) (fx- k 1)])
		   ((fx= k 0))
		 (display "   " out) ) ) )
	  (write-char #\space out)
	  (display (justify (ref bv a) 2 16 #\0) out) )
	(write-char #\space out)
	(do ([j 0 (fx+ j 1)]
	     [a a (fx+ a 1)] )
	    ((or (fx>= j 16) (fx>= a len)))
	  (let ([c (ref bv a)])
	    (if (and (fx>= c 32) (fx< c 128))
		(write-char (integer->char c) out)
		(write-char #\. out) ) ) ) 
	(write-char #\newline out) ) ) ) )


;;; Start interpreting:

(define (run)
  (let* ([extraopts (parse-option-string (or (getenv "CSI_OPTIONS") ""))]
	 [args (cdr (argv))]
	 [loadlater #f]
	 [setup (member "-setup" args)]
	 [kwstyle (member "-keyword-style" args)]
	 [script (or (member "-script" args)
		     (member "-script-meta" args) ) ] )
    (cond [script
	   (unless (pair? (cdr script)) (##sys#error "missing argument to `-script' option"))
	   (command-line-arguments (cddr script))
	   (register-feature! 'script)
	   (set-cdr! (cdr script) '()) 
	   (when (memq (software-type) '(windows msdos))
	     (and-let* ([sname (lookup-script-file (cadr script))])
	       (set-car! (cdr script) sname) ) )
	   (when (string=? "-script-meta" (car script))
	     (set! loadlater (open-input-file (cadr script)))
	     (read-line loadlater)
	     (set! args (append (parse-option-string (read-line loadlater)) args))
	     (set-cdr! script '()) ) ]
	  [else
	   (set! args (append extraopts args))
	   (and-let* ([p (member "--" args)])
	     (set-cdr! p '()) ) ] )
    (let* ([batch (or (member "-batch" args) script setup)]
	   [quiet (or (member "-quiet" args) script setup)]
	   [srfi7 (member "-srfi-7" args)]
	   [r5rs (member "-r5rs" args)]
	   [strict (or r5rs (member "-strict" args))]
	   [strictr (member "-strict-reader" args)]
	   [strict-srfi0 (member "-strict-srfi-0" args)]
	   [strict-letrec (or strict strict-srfi0 (member "-strict-letrec" args))]
	   [hygienic (or r5rs (member "-hygienic" args))]
	   [home (chop-separator (or (getenv "CHICKEN_HOME") installation-home ""))]
	   [ipath (map chop-separator (string-split (or (getenv "CHICKEN_INCLUDE_PATH") "") ";"))] )
      
      (define (collect-options opt)
	(let loop ([opts args])
	  (cond [(member opt opts) 
		 => (lambda (p)
		      (if (null? (cdr p))
			  (##sys#error "missing argument to command-line option" opt)
			  (cons (cadr p) (loop (cddr p)))) ) ]
		[else '()] ) ) )

      (define (loadinit)
	(let* ([prefix (chop-separator (or (getenv "HOME") "."))]
	       [fn (string-append prefix (string pathname-directory-separator) init-file)] )
	  (when (file-exists? fn) 
	    (load fn) ) ) )

      (when (member "-help" args)
	(print-usage)
	(exit 0) )
      (when (or (member "-setup-help" args) (member "-help-setup" args))
	(setup-usage) )
      (when (member "-version" args)
	(print-banner)
	(exit 0) )
      (when (member "-no-warnings" args)
	(unless quiet (display "; Warnings are disabled\n"))
	(set! ##sys#warnings-enabled #f) )
      (unless quiet
	(load-verbose #t)
	(print-banner) )
      (when (member "-case-insensitive" args)
	(unless quiet (display "; Identifiers and symbols are case insensitive\n"))
	(register-feature! 'case-insensitive)
	(case-sensitive #f) )
      (for-each register-feature! (collect-options "-feature"))
      (for-each unregister-feature! (collect-options "-no-feature"))
      (set! ##sys#include-pathnames 
	(delete-duplicates
	 (append (map chop-separator (collect-options "-include-path"))
		 ##sys#include-pathnames
		 ipath
		 (if home (list home) '()) )
	 string=?) )
      (unless hygienic (provide 'match))
      (cond [hygienic
	     (unless quiet
	       (display "; Using hygienic macros")
	       (cond [strict (display " - strict mode\n")]
		     [strict-srfi0 (display " - strict mode with SRFI-0\n")]
		     [else (newline)] ) )
	     (##syncase#install-macro-package (or strict strict-srfi0) (not strict) strict-letrec)
	     (##syncase#install-macro-defs trace-hl-macros) ]
	    [strict
	     (unless quiet (display "; Strict mode\n"))
	     (set! ##sys#strict-mode #t)
	     (unregister-feature! #:srfi-8 #:srfi-2 #:srfi-0 #:srfi-16 #:srfi-9 #:srfi-26)
	     (register-feature! #:strict)
	     (##sys#undefine-non-standard-macros '()) ]
	    [strict-srfi0
	     (unless quiet (display "; Strict mode with SRFI-0\n"))
	     (set! ##sys#strict-mode #t)
	     (unregister-feature! #:srfi-8 #:srfi-2 #:srfi-16 #:srfi-9 #:srfi-26)
	     (register-feature! #:strict)
	     (##sys#undefine-non-standard-macros '(cond-expand)) ] )
      (when (or strict strictr strict-srfi0)
	(strict-reader #t)
	(case-sensitive #f) )
      (when (and strict-letrec (not hygienic))
	(##sys#enable-strict-letrec) )
      (when kwstyle
	(cond [(not-pair? (cdr kwstyle)) 
	       (##sys#error "missing argument to `-keyword-style' option") ]
	      [(string=? "prefix" (cadr kwstyle))
	       (keyword-style #:prefix) ]
	      [(string=? "suffix" (cadr kwstyle))
	       (keyword-style #:suffix) ] ) )
      (when (member "-slib" args)
	(unless slibpath
	  (##sys#error "you need to set the SCHEME_LIBRARY_PATH environment variable") )
	(load (make-pathname slibpath "chicken-slib")) )
      (unless (or (member "-no-init" args) script setup) (loadinit))
      (if setup
	  (if (##sys#provided? 'chicken-setup)
	      (match (cdr setup)
		[() (initialize-setup-stuff)]
		[("-init") (initialize-registry!)]
		[(ext . args) (run-setup-script ext args)]
		[_ (##sys#error "invalid command-line format for `-setup' option")] )
	      (begin
		(print "Sorry, the `-setup' option is not available on this platform.")
		(exit 1) ) )
	  (do ([args args (cdr args)])
	      ((null? args)
	       (when loadlater (load loadlater))
	       (unless batch (read-eval-print-loop)) )
	    (let* ([arg (car args)]
		   [len (string-length arg)] )
	      (cond ((member 
		      arg 
		      '("--" "-batch" "-quiet" "-no-init" "-hygienic" "-no-warnings" "-script"
			"-strict-reader" "-script-meta" "-strict-letrec" "-slib" "-r5rs"
			"-case-insensitive" "-srfi-7" "-strict" "-strict-srfi-0") ) )
		    ((member arg '("-feature" "-no-feature" "-include-path" "-keyword-style"))
		     (set! args (cdr args)) )
		    ((string=? "-visit" arg)
		     (visit (cadr args))
		     (set! args (cdr args)) )
		    ((string=? "-eval" arg)
		     (let ([in (open-input-string (cadr args))])
		       (do ([x (read in) (read in)])
			   ((eof-object? x))
			 (eval x) )
		       (set! args (cdr args)) ) )
		    ((and (>= len 2) (string=? "-:" (substring arg 0 2))))
		    (srfi7
		     (eval 
		      (##sys#expand-srfi-7-program
		       (chop-separator arg)
		       (with-input-from-file arg read)
		       ##sys#error) ) )
		    (else (load arg) ) ) ) ) ) ) ) )

(run)
